<?php

namespace Drupal\config_sync;

use Drupal\config_snapshot\ConfigSnapshotStorageTrait;
use Drupal\config_sync\Plugin\SyncConfigCollectorInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Extension\Extension;

/**
 * The ConfigSyncSnapshotter provides helper functions for taking snapshots of
 * extension-provided configuration.
 */
class ConfigSyncSnapshotter implements ConfigSyncSnapshotterInterface {

  use ConfigSnapshotStorageTrait;
  use ConfigSyncExtensionsTrait;

  /**
   * The app root for the current operation.
   *
   * @var string
   */
  protected $root;

  /**
   * The configuration collector.
   *
   * @var \Drupal\config_sync\Plugin\SyncConfigCollectorInterface
   */
  protected $configCollector;

  /**
   * The provider configuration storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $providerStorage;

  /**
   * The configuration manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * The active configuration store.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $activeStorage;

  /**
   * Constructs a ConfigSyncSnapshotter object.
   *
   * @param string $root
   *   The app root.
   * @param \Drupal\config_provider\Plugin\SyncConfigCollectorInterface $config_collector
   *   The config collector.
   * @param \Drupal\Core\Config\StorageInterface $provider_storage
   *   The provider configuration storage.
   * @param \Drupal\Core\Config\ConfigManagerInterface $config_manager
   *   The configuration manager.
   * @param \Drupal\Core\Config\StorageInterface $active_storage
   *   The active configuration store where the list of enabled modules and
   *   themes is stored.
   */
  public function __construct($root, SyncConfigCollectorInterface $config_collector, StorageInterface $provider_storage, ConfigManagerInterface $config_manager, StorageInterface $active_storage) {
    $this->root = $root;
    $this->configCollector = $config_collector;
    $this->providerStorage = $provider_storage;
    $this->configManager = $config_manager;
    $this->activeStorage = $active_storage;
  }

  /**
   * {@inheritdoc}
   */
  public function refreshExtensionSnapshot($type, array $names, $mode) {
    foreach ($names as $name) {
      $extensions = [];
      $pathname = $this->drupalGetFilename($type, $name);
      $extensions[$name] = new Extension($this->root, $type, $pathname);
      $snapshot_storage = $this->getConfigSnapshotStorage(ConfigSyncSnapshotterInterface::CONFIG_SNAPSHOT_SET, $type, $name);

      switch ($mode) {
        // On install, snapshot configuration in two stages. First, snapshot
        // unaltered configuration. Then, below, apply alters.
        case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL:
          $this->configCollector->addConfigForSnapshotting($extensions);
          break;

        // On import, snapshot fully altered configuration.
        case ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_IMPORT:
          $this->configCollector->addInstallableConfig($extensions);
          break;
      }

      $this->configManager->createSnapshot($this->providerStorage, $snapshot_storage);

      // Conditionally alter the previously added configuration.
      if ($mode === ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL) {
        $this->configCollector->alterConfigSnapshots($extensions);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createSnapshot() {
    $extension_names = $this->getSyncExtensions();

    foreach ($extension_names as $type => $names) {
      if ($type === 'module') {
        $names = $this->listModulesInDependencyOrder($names);
      }
      $this->refreshExtensionSnapshot($type, $names, ConfigSyncSnapshotterInterface::SNAPSHOT_MODE_INSTALL);
    }
  }

  /**
   * Copies the content of one storage to another.
   *
   * @param \Drupal\Core\Config\StorageInterface $source_storage
   *   The storage to synchronize configuration from.
   * @param \Drupal\Core\Config\StorageInterface $snapshot_storage
   *   The storage to synchronize configuration to.
   */
  protected function copyStorage($source_storage, $snapshot_storage) {
    foreach ($source_storage->listAll() as $name) {
      $snapshot_storage->write($name, $source_storage->read($name));
    }
    // Copy collections as well.
    foreach ($source_storage->getAllCollectionNames() as $collection) {
      $source_collection = $source_storage->createCollection($collection);
      $snapshot_collection = $snapshot_storage->createCollection($collection);
      foreach ($source_collection->listAll() as $name) {
        $snapshot_collection->write($name, $source_collection->read($name));
      }
    }
  }

  /**
   * Returns a list of specified modules sorted in order of dependency.
   *
   * @param string[] $module_list
   *   An array of module names.
   *
   * @return string[]
   *   An array of module names.
   */
  protected function listModulesInDependencyOrder($module_list) {
    $module_list = array_combine($module_list, $module_list);

    // Get a list of modules with dependency weights as values.
    $module_data = system_rebuild_module_data();
    // Set the actual module weights.
    $module_list = array_map(function ($module) use ($module_data) {
      return $module_data[$module]->sort;
    }, $module_list);

    // Sort the module list by their weights (reverse).
    arsort($module_list);
    return array_keys($module_list);
  }

  /**
   * Wrapper for drupal_get_filename().
   *
   * @param $type
   *   The type of the item; one of 'core', 'profile', 'module', 'theme', or
   *   'theme_engine'.
   * @param $name
   *   The name of the item for which the path is requested. Ignored for
   *   $type 'core'.
   *
   * @return string
   *   The path to the requested item or an empty string if the item is not
   *   found.
   */
  protected function drupalGetFilename($type, $name) {
    return drupal_get_filename($type, $name);
  }

}
